/* ******************************************************************
 *
 *   version  who    when      what
 *     1.7     Eu   060905     Erste Version mit XML-Import/Export
 *     1.75    Eu   070419     Verbesserte Speicherung des Pfades
 */

import java.util.*;
import java.io.*;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.border.Border;

import javax.xml.parsers.DocumentBuilder; 
import javax.xml.parsers.DocumentBuilderFactory;  
import javax.xml.parsers.FactoryConfigurationError;  
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.*;
import javax.xml.transform.dom.DOMSource; 
import javax.xml.transform.stream.StreamResult; 

import org.xml.sax.SAXException;  
import org.xml.sax.SAXParseException;  

import org.w3c.dom.*;

public class Brett extends JFrame implements ActionListener {
    static final String version = "1.75 April 2007";
    static final String hilfeText = "Eine Beschreibung findet man unter www.fh-friedberg.de/users/euler/tools/springer.htm";
    static final int  SINGLE   = 0;
    static final int  COMPUTER = 1;

    public static String spielerIcon   =  "springer.jpg";
    public static String computerIcon  =  "mndlogo3_mini.gif";
    public static String sperrFeldIcon =  "fels.jpg";
    public static String spielIcon     =  "springer.jpg";

    static int verbose = 0;

    int zeilen  = 8;
    int spalten = 12;
    int size    = 45;
    JButton current;    // Last square
    JLabel[] scoreLabel;
    Spieler[] spieler = new Spieler[2];
    Feld[][] felder;
    boolean gameOver     = false;
    JPanel board;
    JPanel scorePanel;
    FbMenue fbm;
    double sperrFeldProb = 0.05;
    int mode             = COMPUTER;
    int maxValue         = 8;
    int totalValue       = 0;
    String propertyFile  = "springer.ini"; 
    String saveFile      = "save.xml";

    // 
    boolean showPath     = false;
    Color pathColor      = java.awt.Color.blue;
    float pathWidth      = 2.5f;
    ArrayList<JButton> path = new ArrayList(); 
    int[] xPoints;
    int[] yPoints;
 
    public int    getZeilen()  { return zeilen; }
    public int    getSpalten() { return spalten; }
    public void   setZeilen(int z)  { zeilen = z; }
    public void   setSpalten(int s) { spalten = s; }
    public double getSperrFeldProb() { return sperrFeldProb; }
    public void   setSperrFeldProb( double p)  { sperrFeldProb = p; }
    public int    getMode() { return mode; }
    public void   setMode(int m)  { mode = m; }


    public Brett() {
	Properties props = new Properties();
        try {
            props.load( new FileInputStream( propertyFile ) );
	    spielIcon  = 
		props.getProperty("spielIcon", spielIcon);
	    spielerIcon  = 
		props.getProperty("spielerIcon", spielerIcon);
	    computerIcon  = 
		props.getProperty("computerIcon", computerIcon);
	    sperrFeldIcon  = 
		props.getProperty("sperrFeldIcon", sperrFeldIcon);

	    zeilen   = Integer.parseInt((String) props.getProperty("zeilen", "8").trim() );
	    spalten  = Integer.parseInt((String) props.getProperty("spalten", "12").trim());
	    int p    = Integer.parseInt((String) props.getProperty("sperrfeld", "95").trim() );
	    sperrFeldProb = p / 100.;

	    if( props.getProperty("background") != null ) {
		Feld.setBackgroundColor
		    ( MyColor.fromName( props.getProperty("background") ) );
	    }
	    if( props.getProperty("foreground") != null ) {
		Feld.setForegroundColor
		    ( MyColor.fromName( props.getProperty("foreground") ) );
	    }
	    if( props.getProperty("backgroundCW") != null ) {
		Feld.setBackgroundColorCW
		    ( MyColor.fromName( props.getProperty("backgroundCW") ) );
	    }
	    if( props.getProperty("backgroundCB") != null ) {
		Feld.setBackgroundColorCB
		    ( MyColor.fromName( props.getProperty("backgroundCB") ) );
	    }
       
	    pathWidth  = Float.parseFloat((String) props.getProperty("pathWidth", ""+pathWidth).trim() );
	    if( props.getProperty("pathColor") != null ) {
		pathColor = MyColor.fromName( props.getProperty("pathColor") );
	    }
       
        } catch( Exception e ) {
            System.out.println( "Exception: " + e);
        } 

        setTitle( "Springer-Reise");
        setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
	createPlayer();
        initComponents();
	path.add( current );
    }

    /** ******************************************************************
     *  
     */
    void  createPlayer(){
 	if( mode == SINGLE ) {
	    spieler = new Spieler[1];
	    spieler[0] = new Spieler( 0, 0 );
	} else if( mode == COMPUTER ) {
	    spieler = new Spieler[2];
	    spieler[0] = new Spieler( 0, 0 );
	    spieler[1] = new Computer( spalten - 1, zeilen - 1, this );
	}
   }

    /** ******************************************************************
     *  
     */
    private void initComponents() {
        
	fbm = new FbMenue( this );
	setJMenuBar( fbm.getMenuBar() );
 
        board = new JPanel( new GridLayout(zeilen, spalten ));
	fillBoard();
	board.setPreferredSize( new Dimension( spalten * size, zeilen * size ) );

        getContentPane().setLayout( new FlowLayout() );
        getContentPane().add( board );

	JPanel rechts = new JPanel(new GridLayout(0,1), false);
	rechts.add( new JLabel(  new ImageIcon( spielIcon ) ) );

	scorePanel = new JPanel(new GridLayout(0,1), false);
	createScoreLabels();
	rechts.add( scorePanel );

	getContentPane().add( rechts );
        pack();
    }
    
    /** ******************************************************************
     *  
     */
    void createScoreLabels() {
	scoreLabel = new JLabel[spieler.length];
	for( int is=0; is<spieler.length; is++ ) {
	    JLabel label = new JLabel( spieler[is].getName(), JLabel.CENTER);
	    
	    scoreLabel[is] = new JLabel( "0000", JLabel.CENTER );
	    Border  raisedbevel, loweredbevel, twolevel;
	    raisedbevel  = BorderFactory.createRaisedBevelBorder();
	    loweredbevel = BorderFactory.createLoweredBevelBorder();
	    twolevel     =  BorderFactory.createCompoundBorder( raisedbevel, loweredbevel);
	    scoreLabel[is].setBorder( twolevel );
	    
	    scorePanel.add( label );
	    scorePanel.add( scoreLabel[is] );
	    scoreLabel[is].setText( "" + spieler[is].getScore( ) );

	}
    }

    /** ******************************************************************
     *  Flle Brett.
     *  Verwende  optionale Sperrfelder
     *  Im Zwei-Spieler Modus werden Felder mit zuflligen Werten belegt
     */
    void fillBoard() {
	xPoints = new int[zeilen * spalten];
	yPoints = new int[zeilen * spalten];

        int count = 0;
	totalValue = 0;
	felder = new Feld[zeilen][spalten];
        for( int j=0; j<zeilen; j++ ) {
            for( int i=0; i<spalten; i++ ){
		Feld l = null; 
		if( Math.random() > (1.-sperrFeldProb) & istFrei( j, i )  ) {
		    l = new SperrFeld();
		    l.setWert( -10 );
		} else {
		    l = new Feld( "" );
		    if( mode == SINGLE  ) {
			l.setWert( 1 );
		    } else {
			l.setWert();
			totalValue += l.getWert();
		    }
		    if( (i + j%2) % 2 == 0 ) {
			l.setBlack();
		    } else {
			l.setWhite();
		    }
		    l.addActionListener(this);
		    l.setActionCommand( ""+i+" "+j);
		    for( int is=0; is<spieler.length; is++ ) {
			if( i == spieler[is].getX() & j == spieler[is].getY() ) {
			    l.changeColor( );
			    l.setText( null );
			    l.setIcon( spieler[is].getIcon() );
			    l.setWert( -1 );
			    if( is == 0 ) {
				current = l;
			    }
			}
		    }
		    l.addMouseListener
			( new MouseAdapter() {
				public void mouseEntered(MouseEvent e) {
				    smartUpdate( e );
				}
				public void mouseExited(MouseEvent e) {
				    smartUpdate( e );
				}
			    } );
		    
		}
		++count;
		board.add( l );
		felder[j][i] = l;
            }
            ++count;
        }
    }


    /** ******************************************************************
     * Repaint only the area of the Button
     */
     void  smartUpdate( MouseEvent e ) {
	Rectangle r = ((JButton ) e.getSource()).getBounds();
	r.x += board.getX() + getInsets().left;
	r.y += board.getY() + fbm.getMenuBar().getHeight() + getInsets().top;
	repaint( r.x, r.y, r.width, r.height );
    }

    boolean istFrei( int i, int j ) {
	for( int is=0; is<spieler.length; is++ ) {
	    if( i == spieler[is].getY() & j == spieler[is].getX() ) {
		return false;
	    }
	}
	return true;
    }

    public int getWert( int i, int j ) {
	//System.out.println( i + " " + j);
	if( i < 0 | j < 0 | i >= zeilen | j >= spalten ) return -2;
	if( ! felder[i][j].isEnabled() ) return -3;
	if( ! istFrei( i, j ) )  return -4;
	return felder[i][j].getWert();
    }
    
    public void move( int i, int j, Computer c ) {
	felder[i][j].changeColor( );
	felder[i][j].setText( null );
	felder[i][j].setIcon( c.getIcon()  ); 
	felder[c.getY()][c.getX()].setEnabled( false );
    }

    /** ******************************************************************
     * 
     */
    public void paint(Graphics g) {
	super.paint(g);
	//super.paintComponent(g); 
	if( showPath ) {
	    Graphics2D g2 = (Graphics2D) g;
	    g2.setColor( pathColor );
	    g2.setStroke( new BasicStroke( pathWidth ) );
	    int count = 0;
	    int xoff = board.getX() + getInsets().left;
	    int yoff =  board.getY() 
		+ fbm.getMenuBar().getHeight() + getInsets().top;
	    for( JButton b : path ) {
		xPoints[count] = b.getX() + xoff + b.getWidth() / 2;
		yPoints[count] = b.getY() + yoff + b.getHeight() / 2;
		++count;
	    }
	    g2.drawPolyline( xPoints, yPoints, count );
	}
    }

    /** ******************************************************************
     *  Event-Handler fuer alle Menues
     */
    public void actionPerformed(ActionEvent e) {
        String cmd = e.getActionCommand();
        // System.out.println( cmd );

	if( cmd.equals( "H_ber" ) ) {
	    JOptionPane.showMessageDialog(this, "FH-FB Springerspiel, Version " + version); 
	} else if( cmd.equals( "H_Hilfe" ) ) {
	    JOptionPane.showMessageDialog(this, hilfeText ); 

	} else if( cmd.equals( "G_Speichern" ) ) {
	    JOptionPane.showMessageDialog(this, "Speichern in Datei " + saveFile); 
	    saveXML();
	    repaint();

	} else if( cmd.equals( "G_Laden" ) ) {
	    JOptionPane.showMessageDialog
		( this, 
		  "Laden von Datei <" 
		  + saveFile 
		  +">. ACHTUNG! Unvollstndige Implementierung." ); 
	    loadXML();
	    
	    mode = SINGLE;
	    newStart();

	} else if( cmd.equals( "G_Neu" ) ) {
	    System.out.println( "Neues Spiel" );
	    NGDialog gd = new NGDialog( this, "test" );
	    gd.setVisible(true);
	    createPlayer();
	    newStart();

	} else if( cmd.equals( "G_Aufgeben" ) ) {
	    spielerGibtAuf();

	} else if( cmd.equals( "G_Exit" ) ) {
	    System.exit(0);

	} else if( cmd.equals( "P_Ein/Aus" ) ) {
	    showPath  = ! showPath;
	    repaint();

	} else { 
	    if( gameOver ) return;
	    StringTokenizer st = new StringTokenizer( cmd );
	    int x = -1, y = -1;
	    try {
		x = Integer.parseInt(st.nextToken());
		y = Integer.parseInt(st.nextToken());
	    } catch ( Exception eparse ) {
		System.out.println( "Unbekannter Befehl: " + cmd );
		return;
	    }
	    if( mode == COMPUTER ) {
		if( x == spieler[1].getX() & y == spieler[1].getY() ) {
		    return;
		}
	    }

	    // System.out.println( x + "  " + y );
	    int xd = Math.abs( x - spieler[0].getX() );
	    int yd = Math.abs( y - spieler[0].getY() );
        
	    if( xd == 2 & yd == 1 | xd == 1 & yd == 2 ) {
		spieler[0].setX( x );
		spieler[0].setY( y );
		Feld b = (Feld) e.getSource();
		
		path.add( b );
	
		repaint();

		b.ausfhren(  spieler[0] );
		scoreLabel[0].setText( "" + spieler[0].incScore( b.getWert() ) );
		current.setEnabled( false );
		current = b;
		if( mode == COMPUTER ) {
		    ((Computer) spieler[1]).move();
		    scoreLabel[1].setText( "" + spieler[1].getScore( ) );
		}
	    }
        }
    }
    
    /** ******************************************************************
     *  Computer kann nicht mehr ziehen und gibt auf
     */
    public void newStart() 
    {
	board.removeAll();
	path.clear();
	board.setLayout( new GridLayout(zeilen, spalten ));
	fillBoard();
	board.setPreferredSize( new Dimension( spalten * size, zeilen * size ) );
	board.repaint();
	scorePanel.removeAll();
	createScoreLabels();

	pack();
	repaint();

	path.add( current );

	gameOver = false;
    }

    /** ******************************************************************
     *  Computer kann nicht mehr ziehen und gibt auf
     */
    public void computerGibtAuf() 
    {
	int bonus = totalValue - spieler[0].getScore() - spieler[1].getScore();
	JOptionPane.showMessageDialog(this, "Computer gibt auf!\nBonus " + bonus + " Punkte" ); 
	scoreLabel[0].setText( "" + (spieler[0].getScore( ) + bonus ) );
	gameOver = true;
    }

    /** ******************************************************************
     *  
     */
    public void spielerGibtAuf() 
    {
	if( mode == COMPUTER ) {
	    int bonus = totalValue - spieler[0].getScore() - spieler[1].getScore();
	    JOptionPane.showMessageDialog(this, "Spieler gibt auf!\nBonus " + bonus + " Punkte" ); 
	    scoreLabel[1].setText( "" + (spieler[1].getScore( ) + bonus ) );
	}
	gameOver = true;
    }

    /** ******************************************************************
     *  Laden des Spielstandes aus XML-Datei <BR>
     */
    void loadXML() {
        Document document = null;
	DocumentBuilderFactory factory =
            DocumentBuilderFactory.newInstance();
        try {
           DocumentBuilder builder = factory.newDocumentBuilder();
           document = builder.parse( new File( saveFile ) );
	   
        } catch (SAXException sxe) {
	    // Error generated during parsing)
	    Exception  x = sxe;
	    if (sxe.getException() != null) {
		x = sxe.getException();
	    }
	    x.printStackTrace();

        } catch (ParserConfigurationException pce) {
            // Parser with specified options can't be built
            pce.printStackTrace();

        } catch (IOException ioe) {
           // I/O error
           ioe.printStackTrace();
        }

	NodeList nl = document.getChildNodes();
	for( int i=0; i<nl.getLength(); i++ ) {
	    Node n = nl.item( i );
	    System.out.println( "child node # " + i + ": " +n.getNodeName() );
	}
	
	Element root =  document.getDocumentElement();
	System.out.println( "root: " + root.getNodeName() );

	NodeList nlroot = root.getChildNodes( );
	for( int k=0; k<nlroot.getLength(); k++ ) {
	    Node nvc = nlroot.item(k);
	    if( nvc.getNodeType() == Node.ELEMENT_NODE ) {
		//System.out.println( "Element: " );
		Element e = (Element) nvc;
		String tag =  e.getTagName();
		if( tag.equals( "Breite" ) ) {
		    spalten = Integer.parseInt( e.getTextContent() );
		    System.out.println( "Breite: " + spalten );
		} else 	if( tag.equals( "Hoehe" ) ) {
		    zeilen = Integer.parseInt( e.getTextContent() );
		    System.out.println( "Zeilen: " + zeilen );
		} else 	if( tag.equals( "Spieler" ) ) {
		    if( verbose > 0 ) System.out.println("********* Spieler ***************");
		    NodeList nlspieler = nvc.getChildNodes( );
		    for( int l=0; l<nlspieler.getLength(); l++ ) {
			Node n = nlspieler.item(l);
			if( n.getNodeType() == Node.ELEMENT_NODE ) {
			    Element e2 = (Element) n;
			    String tag2 =  e2.getTagName();
			    if( verbose > 0 ) {
				System.out.println( tag2 );
			    }
			    if( tag2.equals( "Name" ) ) {
				System.out.println( "Name: " + e2.getTextContent() );
				spieler[0].setName( e2.getTextContent() );
			    } else if( tag2.equals( "Position" ) ) {
				spieler[0].setX
				    ( Integer.parseInt
				      ( e2.getAttribute( "Spalte" ) ) );
				spieler[0].setY
				    ( Integer.parseInt
				      ( e2.getAttribute( "Zeile" ) ) );
			    } else if( tag2.equals( "Punkte" ) ) {
				spieler[0].setScore( Integer.parseInt( e2.getTextContent() ) );
			    }
			}
		    }
		}
	    }
	}
	
    }

    /** ******************************************************************
     *  Speichern des aktuellen Spielstandes in XML-Datei <BR>
     *  
     */
    void saveXML() {
	String css = "type=\"text/css\" href=\"springer.css\"";
	Document document = null;
	DocumentBuilderFactory factory =
	    DocumentBuilderFactory.newInstance();

	try {
	    DocumentBuilder builder =  factory.newDocumentBuilder();
	    document = builder.newDocument(); 
	    document.appendChild
		( document.createComment(" Springer-Reise " + 
				Calendar.getInstance().getTime().toString() +" ") 
		  );
	    document.appendChild(
			 document.createProcessingInstruction("xml-stylesheet", css)
			 );
			 
	    Element root = 
		(Element) document.createElement("Springerspiel"); 
	    document.appendChild(root);
	    Element breite = (Element) document.createElement("Breite");
	    breite.appendChild( document.createTextNode( ""+ spalten  ) );
	    root.appendChild( breite );
			     
	    Element hoehe = (Element) document.createElement("Hoehe");
	    hoehe.appendChild( document.createTextNode( ""+ zeilen ) );
	    root.appendChild( hoehe );
			     
	    Element spielerE = (Element) document.createElement("Spieler");
	    Element name = (Element) document.createElement("Name");
	    name.appendChild( document.createTextNode( spieler[0].getName() ) );
	    spielerE.appendChild( name );

	    Element position = (Element) document.createElement("Position");
	    position.setAttribute("Spalte", ""+ spieler[0].getX()  );
	    position.setAttribute("Zeile",  ""+ spieler[0].getY()  );
	    spielerE.appendChild( position );
			     
	    Element punkte = (Element) document.createElement("Punkte");
	    punkte.appendChild( document.createTextNode( ""+ spieler[0].getScore() ) );
	    spielerE.appendChild( punkte );
	    
	    root.appendChild( spielerE );
			     
	    File f = new File( saveFile );
	    
	    // Use a Transformer for output
	    TransformerFactory tFactory =
		TransformerFactory.newInstance();
	    
	    Transformer transformer = tFactory.newTransformer();
	    transformer.setOutputProperty(OutputKeys.METHOD, "xml");
	    transformer.setOutputProperty(OutputKeys.INDENT, "yes");
	    transformer.setOutputProperty(OutputKeys.ENCODING, "ISO-8859-1"); 

	    DOMSource source = new DOMSource(document);
	    StreamResult result = new StreamResult( new FileOutputStream( f ) );
	    transformer.transform(source, result); 
	    transformer.transform(source, new StreamResult( System.out )); 
	    
	} catch (Exception ex) {
	    ex.printStackTrace();
	}
    } 

    // *********************************************************
    static void printHelp() {
	System.out.println("USAGE: " + " [ arg1 arg2 ... ] input");
 	System.out.println("  -v        :  verbose mode ");
 	System.out.println("  -p        :  print grid for latex ");
    }

    /** ******************************************************************
     *
     */
    public static void main(String[] args) {
	
	Brett brett = new Brett();

 	for( int i=0; i<args.length; i++ ) {
	    if( args[i].equals("-h") ) {
		printHelp();
		System.exit(0);
	    } else if( args[i].equals("-i") ) {
		brett.saveFile = args[++i];
	    } else if( args[i].equals("-v") ) {
		brett.verbose = 1;
	    }
	}
	
	brett.show();
    }
}
